"""
This script largely controls the KiCAD CLI to generate
gerbers/drill files/etc... required to make circuit boards
"""

import logging
import re
import subprocess
import sys
import zipfile
from functools import cache
from os import getenv, PathLike
from pathlib import Path
from time import time
from typing import Optional

import git
import semver

import atopile.errors
from atopile import config

log = logging.getLogger(__name__)


@cache
def get_cli_version(path: Path) -> semver.Version:
    """Get the version of the KiCAD CLI."""
    version_output = subprocess.check_output(
        [str(path), "--version"],
        text=True,
        # this timeout generally feels excessive, but with security it can happen
        timeout=5,
    )
    return semver.Version.parse(version_output)


@cache
def find_kicad_cli() -> PathLike:
    """Figure out what to call for the KiCAD CLI."""
    if sys.platform.startswith("darwin"):
        kicad_cli_candidates = list(Path("/Applications/KiCad/").glob("**/kicad-cli"))

        def _get_cli_version(path: Path) -> semver.Version:
            try:
                return get_cli_version(path)
            except (ValueError, atopile.errors.AtoError):
                return semver.Version("0.0.0")

        best_candidate = max(kicad_cli_candidates, key=_get_cli_version)
        return best_candidate
    elif sys.platform.startswith("linux"):
        return "kicad-cli"  # assume it's on the PATH
    elif sys.platform.startswith("win"):
        kicad_cli_candidates = list(
            (Path(getenv("ProgramFiles")) / "KiCad").glob("**/kicad-cli.exe"))

        def _get_cli_version(path: Path) -> semver.Version:
            try:
                return get_cli_version(path)
            except (ValueError, atopile.errors.AtoError):
                return semver.Version("0.0.0")

        best_candidate = max(kicad_cli_candidates, key=_get_cli_version)
        return best_candidate


def run(*args, timeout_time: Optional[float] = 10, **kwargs) -> None:
    """Run a subprocess"""
    process = subprocess.Popen(
        *args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, **kwargs
    )

    def _do_logging():
        outs, errs = process.communicate(timeout=0.1)
        for line in outs.splitlines():
            log.info(line)
        for line in errs.splitlines():
            log.error(line)

    start_time = time()
    while timeout_time is not None and time() - start_time < timeout_time:
        if process.poll() is not None:
            break
        try:
            _do_logging()
        except subprocess.TimeoutExpired:
            continue
    else:
        process.kill()

    exit_code = process.wait()
    _do_logging()

    if exit_code != 0:
        raise atopile.errors.AtoError(
            f"Command {args} failed with exit code {exit_code}"
        )


@cache
def _get_short_githash(project_path: Path) -> str:
    """Get the short git hash for the project."""
    try:
        repo = git.Repo(project_path, search_parent_directories=True)
    except git.InvalidGitRepositoryError:
        atopile.errors.AtoError("Project is not a git repository").log(
            log, logging.WARNING
        )
        short_githash = "nogit"
    else:
        short_githash_length = 7
        if repo.is_dirty():
            short_githash = "dirty-" + repo.head.commit.hexsha[:short_githash_length]
            atopile.errors.AtoError(
                f"There are non-commited changes in your repo. Git hash is: '{short_githash}'"
            ).log(log, logging.WARNING)
        else:
            short_githash = repo.head.commit.hexsha[:short_githash_length]

    return short_githash


# FIXME: can't use a regular cached function here because build_ctx is mutable
_ensure_modded_kicad_pcb_cache = {}
def _ensure_modded_kicad_pcb(build_ctx: config.BuildContext) -> Path:
    """Ensure the KiCAD PCB file has been modified for manufacturing."""
    # First, check if the build_ctx is in the cache
    if id(build_ctx) in _ensure_modded_kicad_pcb_cache:
        return _ensure_modded_kicad_pcb_cache[id(build_ctx)]

    # If there's no layout, we can't generate manufacturing data
    if not build_ctx.layout_path:
        atopile.errors.AtoError(
            "Layout must be available to generate manufacturing data"
        ).log(log, logging.WARNING)
        return

    # Ensure the build directory exists
    build_ctx.build_path.mkdir(parents=True, exist_ok=True)

    # Replace constants in the board file
    project_path = config.get_project_context().project_path
    short_githash = _get_short_githash(project_path)

    modded_kicad_pcb = build_ctx.output_base.with_suffix(".kicad_pcb")

    githash_kw = re.compile(re.escape("{{GITHASH}}"))
    with build_ctx.layout_path.open("r") as f_src, modded_kicad_pcb.open("w") as f_dst:
        for line in f_src:
            f_dst.write(githash_kw.sub(short_githash, line))

    # Finally, cache the result
    _ensure_modded_kicad_pcb_cache[id(build_ctx)] = modded_kicad_pcb

    return modded_kicad_pcb


def generate_manufacturing_data(build_ctx: config.BuildContext) -> None:
    """Generate manufacturing data for the project."""
    # If there's no layout, we can't generate manufacturing data
    if not build_ctx.layout_path:
        atopile.errors.AtoError(
            "Layout must be available to generate manufacturing data"
        ).log(log, logging.WARNING)
        return

    # Ensure the build directory exists
    build_ctx.build_path.mkdir(parents=True, exist_ok=True)

    # Replace constants in the board file
    project_path = config.get_project_context().project_path
    short_githash = _get_short_githash(project_path)
    modded_kicad_pcb = _ensure_modded_kicad_pcb(build_ctx)

    # Setup for Gerbers
    gerber_dir = build_ctx.output_base.with_name(
        f"{build_ctx.output_base.name}-gerbers-{short_githash}"
    )
    gerber_dir.mkdir(exist_ok=True, parents=True)
    gerber_dir_str = str(gerber_dir)
    if not gerber_dir_str.endswith("/"):
        gerber_dir_str += "/"

    kicad_cli = find_kicad_cli()

    # Generate Gerbers
    run(
        [
            kicad_cli,
            "pcb",
            "export",
            "gerbers",
            "-o",
            gerber_dir_str,
            str(modded_kicad_pcb),
        ]
    )
    run(
        [
            kicad_cli,
            "pcb",
            "export",
            "drill",
            "-o",
            gerber_dir_str,
            str(modded_kicad_pcb),
        ]
    )

    # Zip Gerbers
    zip_path = gerber_dir.with_suffix(".zip")
    with zipfile.ZipFile(zip_path, "w") as zip_file:
        for file in gerber_dir.glob("*"):
            zip_file.write(file)

    # Position files need some massaging for JLCPCB
    # We just need to replace the first row
    pos_path = build_ctx.output_base.with_suffix(".pos.csv")
    run(
        [
            kicad_cli,
            "pcb",
            "export",
            "pos",
            "--format",
            "csv",
            "--units",
            "mm",
            "--use-drill-file-origin",
            "-o",
            str(pos_path),
            str(modded_kicad_pcb),
        ]
    )
    pos_contents = pos_path.read_text().splitlines()
    pos_contents[0] = "Designator,Value,Package,Mid X,Mid Y,Rotation,Layer"
    pos_path.write_text("\n".join(pos_contents))


def generate_drc_report(build_ctx: config.BuildContext) -> None:
    """Generate manufacturing data for the project."""
    if not build_ctx.layout_path:
        atopile.errors.AtoError(
            "Layout must be available to generate manufacturing data"
        ).log(log, logging.WARNING)
        return

    # Ensure the build directory exists
    build_ctx.build_path.mkdir(parents=True, exist_ok=True)

    # Replace constants in the board file
    modded_kicad_pcb = _ensure_modded_kicad_pcb(build_ctx)

    # Setup for Gerbers
    kicad_cli = find_kicad_cli()
    version = get_cli_version(kicad_cli)

    # Run DRCs, if possible via CLI
    # TODO: pull this out into another target
    if version >= semver.Version(8):
        try:
            run(
                [
                    kicad_cli,
                    "pcb",
                    "drc",
                    "-o",
                    str(build_ctx.build_path / "drc-report.rpt"),
                    "--exit-code-violations",
                    str(modded_kicad_pcb),
                ]
            )
        except atopile.errors.AtoError as e:
            drc_err = atopile.errors.AtoError("DRCs failed")
            if build_ctx.fail_on_drcs:
                raise drc_err from e
            else:
                drc_err.log(log, logging.WARNING)



"""
Utils related to handling the parse tree
"""
from pathlib import Path
from typing import Any, Optional

from antlr4 import (
    InputStream,
    Lexer,
    ParserRuleContext,
    ParseTreeVisitor,
    Token,
)


def get_src_info_from_token(token: Token) -> tuple[str, int, int]:
    """Get the source path, line, and column from a context"""
    input_stream: InputStream = token.getInputStream()
    return input_stream.name, token.line, token.column


def get_src_info_from_ctx(ctx: ParserRuleContext) -> tuple[str | Path, int, int, int, int]:
    """Get the source path, line, and column from a context"""
    token: Token = ctx.start
    _, stop_line, stop_char = get_src_info_from_token(ctx.stop)
    return *get_src_info_from_token(token), stop_line, stop_char


def format_src_info(ctx: ParserRuleContext) -> str:
    """Format the source path, line, and column"""
    src, start_line, start_col, _, _ = get_src_info_from_ctx(ctx)
    return f"{src}:{start_line}:{start_col}"


# FIXME: I hate this pattern
# It should instead at least return a list of tokens
# for processing in a regular for loop
class _Reconstructor(ParseTreeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.txt = ""
        self.last_line = None
        self.last_col = None

    def visitTerminal(self, node) -> str:
        symbol: Token = node.getSymbol()

        if self.last_line is None:
            self.last_line = symbol.line
            self.last_col = symbol.start

        if symbol.line > self.last_line:
            self.txt += "\n" * (symbol.line - self.last_line)
            self.last_col = 0

        self.txt += " " * (symbol.start - self.last_col - 1)

        self.last_line = symbol.line
        self.last_col = symbol.stop

        self.txt += node.getText()
        return super().visitTerminal(node)


def reconstruct(ctx: ParserRuleContext) -> str:
    """Reconstruct the source code from a parse tree"""
    reco = _Reconstructor()
    reco.visit(ctx)
    return reco.txt


def get_comment_from_token(token: Token) -> Optional[str]:
    """Return the comment from a token's start line."""
    lexer: Optional[Lexer] = token.getTokenSource()
    if not lexer or not hasattr(lexer, "comments"):
        return None

    comments: dict[tuple[Any, int], str] = lexer.comments
    line: int = token.line

    if input_stream := token.getInputStream():
        source_name = input_stream.name
    else:
        return None

    return comments.get((source_name, line))



"""
Work with expressions and ranged values.
"""

import collections.abc
from collections import ChainMap
from typing import Callable, Mapping, Optional, Union

import pint
from antlr4 import ParserRuleContext
from attrs import define, frozen
from pint.facets.plain import PlainUnit

from atopile import address, errors

_UNITLESS = pint.Unit("")


def _custom_float_format(value, max_decimals: int):
    """
    Format the float to have up to max_decimals places, but fewer if there's no more precision.
    """
    # Format with fixed-point notation, up to max_decimals
    formatted = f"{value:.{max_decimals}f}"
    # Remove trailing zeros and the decimal point if not needed
    return formatted.rstrip("0").rstrip(".")


_multiplier_map = {
    "femto": "f",
    "pico": "p",
    "nano": "n",
    "micro": "u",
    "milli": "m",
    "": "",
    "kilo": "k",
    "mega": "M",
    "giga": "G",
    "tera": "T",
}


_pretty_unit_map = {
    "": "",  # Dimensionless is desirable too
    "volt": "V",
    "ohm": "Ω",
    "ampere": "A",
    "watt": "W",
    "hertz": "Hz",
    "farad": "F",
    "henry": "H",
    "second": "s",
}


pretty_unit_map = {lm + lu: sm + su for lm, sm in _multiplier_map.items() for lu, su in  _pretty_unit_map.items()}
favorite_units_map = {pint.Unit(k).dimensionality: pint.Unit(k) for k in _pretty_unit_map}


def _convert_to_favorite_unit(qty: pint.Quantity) -> pint.Quantity:
    """Convert the quantity to the favorite unit for its dimensionality."""
    # If there's a favorite unit for this dimensionality, use it
    if qty.units.dimensionality in favorite_units_map:
        qty = qty.to(favorite_units_map[qty.units.dimensionality])

    # Compact the units to standardise them
    qty = qty.to_compact()

    return qty


def _best_units(qty_a: pint.Quantity, qty_b: pint.Quantity) -> PlainUnit:
    """Return the best unit for the two quantities."""
    a_fav = _convert_to_favorite_unit(qty_a)
    b_fav = _convert_to_favorite_unit(qty_b)

    # If converting b to a's units results in a shorter representation, use a's units
    # Otherwise, use b's units
    if len(str(qty_b.to(a_fav.units).magnitude)) < len(str(qty_a.to(b_fav.units).magnitude)):
        return a_fav.units
    return b_fav.units


def pretty_unit(qty: pint.Quantity) -> tuple[float, str]:
    """Return the most favorable magnitude and unit for the given quantity."""
    if qty.units.dimensionless:
        return qty.magnitude, ""

    qty = _convert_to_favorite_unit(qty)

    # Convert the units to a pretty string
    units = str(qty.units)
    pretty_unit = pretty_unit_map.get(units, units)
    return qty.magnitude, pretty_unit


class RangedValue:
    """
    Let's get physical!

    Ranged values are designed to represent a range of physical values, such as
    a voltage or current, including tolerances.
    """

    def __init__(
        self,
        val_a: Union[float, int, pint.Quantity],
        val_b: Optional[Union[float, int, pint.Quantity]] = None,
        unit: Optional[str | PlainUnit | pint.Unit] = None,
        str_rep: Optional[str] = None,
    ):
        # This is a bit of a hack, but simplifies upstream code marginally
        if val_b is None:
            val_b = val_a

        # If we're given a unit, use it. Otherwise, try to infer the unit from the inputs.
        if unit:
            self.unit = pint.Unit(unit)
        elif isinstance(val_a, pint.Quantity) and isinstance(val_b, pint.Quantity):
            self.unit = _best_units(val_a, val_b)
        elif isinstance(val_a, pint.Quantity):
            self.unit = val_a.units
        elif isinstance(val_a, pint.Quantity):
            self.unit = val_b.units
        else:
            self.unit = _UNITLESS

        # If the inputs are pint Quantities, convert them to the same unit
        if isinstance(val_a, pint.Quantity):
            val_a_mag = val_a.to(self.unit).magnitude
        else:
            val_a_mag = val_a

        if isinstance(val_b, pint.Quantity):
            val_b_mag = val_b.to(self.unit).magnitude
        else:
            val_b_mag = val_b

        assert isinstance(val_a_mag, (float, int))
        assert isinstance(val_b_mag, (float, int))

        # Make the noise
        self.str_rep = str_rep
        self.min_val = min(val_a_mag, val_b_mag)
        self.max_val = max(val_a_mag, val_b_mag)

    def to(self, unit: str | PlainUnit | pint.Unit) -> "RangedValue":
        """Return a new RangedValue in the given unit."""
        return RangedValue(self.min_qty, self.max_qty, unit)

    def pretty_str(
        self,
        max_decimals: Optional[int] = 2,
        format_: Optional[str] = None,
    ) -> str:
        """Return a pretty string representation of the RangedValue."""
        def _f(val: float):
            if max_decimals is None:
                return str(val)
            return _custom_float_format(val, max_decimals)

        if self.str_rep and format_ is None:
            return self.str_rep

        # Single-ended
        if self.tolerance == 0 or (
            self.tolerance_pct and
            self.tolerance_pct * 1e4 < pow(10, -max_decimals)
        ):
            nom, unit = pretty_unit(self.nominal * self.unit)
            return f"{_f(nom)}{unit}"

        # Bound values
        if (self.tolerance_pct and self.tolerance_pct > 20) or format_ == "bound":
            min_val, min_unit = pretty_unit(self.min_qty)
            max_val, max_unit = pretty_unit(self.max_qty)

            if min_unit == max_unit or min_val == 0 or max_val == 0:
                return f"{_f(min_val)} to {_f(max_val)} {max_unit}"
            return f"{_f(min_val)}{min_unit} to {_f(max_val)}{max_unit}"

        # Bilateral values
        nom, unit = pretty_unit(self.nominal * self.unit)
        tol, tol_unit = pretty_unit(self.tolerance * self.unit)

        if nom == 0:
            return f"± {_f(tol)}{tol_unit}"
        if unit == tol_unit:
            return f"{_f(nom)} ± {_f(tol)} {unit}"
        return f"{_f(nom)}{unit} ± {_f(tol)}{tol_unit}"

    def __str__(self) -> str:
        return self.pretty_str()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.min_val}, {self.max_val}, '{self.unit}')"
        )

    @property
    def nominal(self) -> float:
        """Return the nominal value of the RangedValue."""
        return (self.min_val + self.max_val) / 2

    @property
    def tolerance(self) -> float:
        """Return the tolerance of the RangedValue."""
        return (self.max_val - self.min_val) / 2

    @property
    def tolerance_pct(self) -> Optional[float]:
        """Return the tolerance as a percentage of the nominal value."""
        if self.nominal == 0:
            return None
        return abs(self.tolerance / self.nominal * 100)

    def to_dict(self) -> dict:
        """Convert the Physical instance to a dictionary."""
        min_qty = _convert_to_favorite_unit(self.min_qty)
        multiplier = min_qty.magnitude / self.min_val
        return {
            "unit": str(min_qty.units),
            "min_val": min_qty.magnitude,
            "max_val": self.max_val * multiplier,
            # TODO: remove these - we shouldn't be duplicating this kind of information
            "nominal": self.nominal * multiplier,
            "tolerance": self.tolerance * multiplier,
            "tolerance_pct": self.tolerance_pct,
        }

    @property
    def min_qty(self) -> pint.Quantity:
        """Return the minimum of self as a pint Quantity."""
        return self.unit * self.min_val

    @property
    def max_qty(self) -> pint.Quantity:
        """Return the maximum of self as a pint Quantity."""
        return self.unit * self.max_val

    @classmethod
    def _ensure(cls, thing) -> "RangedValue":
        if isinstance(thing, RangedValue):
            return thing
        return cls(thing, thing, _UNITLESS)

    def __mul__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        other = self._ensure(other)

        new_values = [
            self.min_qty * other.min_qty,
            self.min_qty * other.max_qty,
            self.max_qty * other.min_qty,
            self.max_qty * other.max_qty,
        ]

        return self.__class__(
            min(new_values),
            max(new_values),
        )

    def __rmul__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self.__mul__(other)

    def __pow__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        if isinstance(other, RangedValue):
            if not (other.unit.dimensionless and other.min_val == other.max_val):
                raise ValueError("Exponent must be a constant valueless quantity")
            other = other.min_val

        return self.__class__(self.min_qty**other, self.max_qty**other)

    @classmethod
    def _do_truediv(
        cls,
        numerator: Union["RangedValue", float, int],
        denominator: Union["RangedValue", float, int],
    ) -> "RangedValue":
        numerator = cls._ensure(numerator)
        denominator = cls._ensure(denominator)

        new_values = [
            numerator.min_qty / denominator.min_qty,
            numerator.min_qty / denominator.max_qty,
            numerator.max_qty / denominator.min_qty,
            numerator.max_qty / denominator.max_qty,
        ]

        return cls(
            min(new_values),
            max(new_values),
        )

    def __truediv__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self._do_truediv(self, other)

    def __rtruediv__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self._do_truediv(other, self)

    def __add__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        other = self._ensure(other)

        return self.__class__(
            self.min_qty + other.min_qty,
            self.max_qty + other.max_qty,
        )

    def __radd__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self.__add__(other)

    def __sub__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        other = self._ensure(other)

        return self.__class__(
            self.min_qty - other.max_qty,
            self.max_qty - other.min_qty,
        )

    def __rsub__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self.__sub__(other)

    def __neg__(self) -> "RangedValue":
        return self.__class__(-self.max_qty, -self.min_qty, self.unit)

    def within(self, other: Union["RangedValue", float, int]) -> bool:
        """Check that this RangedValue completely falls within another."""
        if not isinstance(other, RangedValue):
            if not self.unit.dimensionless:
                raise ValueError(
                    "Can only compare RangedValue to a dimensionless quantity"
                )
            return self.min_val == self.max_val == other
        return self.min_qty >= other.min_qty and other.max_qty >= self.max_qty

    def __lt__(self, other: Union["RangedValue", float, int]) -> bool:
        other = self._ensure(other)
        return self.max_qty < other.min_qty

    def __gt__(self, other: Union["RangedValue", float, int]) -> bool:
        other = self._ensure(other)
        return self.min_qty > other.max_qty

    def __le__(self, other: Union["RangedValue", float, int]) -> bool:
        other = self._ensure(other)
        return self.max_qty <= other.min_qty

    def __ge__(self, other: Union["RangedValue", float, int]) -> bool:
        other = self._ensure(other)
        return self.min_qty >= other.max_qty

    def __eq__(self, other: object) -> bool:
        # NOTE: realistically this is only useful for testing
        if isinstance(other, RangedValue):
            return self.min_qty == other.min_qty and self.max_qty == other.max_qty

        # NOTE: this doesn't work for farenheit or kelvin, but everything else is okay
        if self.min_val == self.max_val == other and (
            self.unit.dimensionless or other == 0
        ):
            return True
        return False

    def __req__(self, other: Union["RangedValue", float, int]) -> bool:
        return self.__eq__(other)

    def __or__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        other = self._ensure(other)

        # make sure there's some overlap
        if self.max_qty < other.min_qty or other.max_qty < self.min_qty:
            raise errors.AtoValueError(f"Ranges ({self}, {other}) do not overlap")

        return self.__class__(
            min(self.min_qty, other.min_qty),
            max(self.max_qty, other.max_qty),
        )

    def __ror__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self.__or__(other)

    def __and__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        other = self._ensure(other)

        # make sure there's some overlap
        if self.max_qty < other.min_qty or other.max_qty < self.min_qty:
            raise errors.AtoValueError(f"Ranges ({self}, {other}) do not overlap")

        return self.__class__(
            max(self.min_qty, other.min_qty),
            min(self.max_qty, other.max_qty),
        )

    def __rand__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self.__and__(other)

    def min(self) -> "RangedValue":
        """Return a new RangedValue with the minimum value."""
        return self.__class__(self.min_qty, self.min_qty)

    def max(self) -> "RangedValue":
        """Return a new RangedValue with the maximum value."""
        return self.__class__(self.max_qty, self.max_qty)


NumericishTypes = Union["Expression", RangedValue, float, int, "Symbol"]


@frozen
class Symbol:
    """Represent a symbol."""

    key: collections.abc.Hashable

    def __call__(self, context: Mapping) -> RangedValue:
        """Return the value of the symbol."""
        thing = context[self.key]
        if callable(thing):
            return thing(context)
        return thing


# TODO: figure out how to pretty print these with the symbols etc...
@define
class Expression:
    """Represent an expression."""

    symbols: set[Symbol]
    lambda_: Callable[[Mapping[str, NumericishTypes]], RangedValue]
    src_ctx: ParserRuleContext | None = None

    @classmethod
    def from_expr(cls, expr: "Expression") -> "Expression":
        """Create an expression from another expression."""
        return cls(symbols=expr.symbols, lambda_=expr.lambda_)

    @classmethod
    def from_numericish(cls, thing: NumericishTypes) -> "Expression":
        """Create an expression from a numericish thing."""
        if isinstance(thing, Expression):
            return cls.from_expr(thing)
        if isinstance(thing, RangedValue):
            return cls(symbols=set(), lambda_=lambda context: thing)
        if isinstance(thing, Symbol):
            return cls(symbols={thing}, lambda_=thing)
        return cls(symbols=set(), lambda_=lambda context: RangedValue(thing, thing))

    def __call__(self, context: Mapping[str, NumericishTypes]) -> RangedValue:
        return self.lambda_(context)

    def substitute(
        self, substitutions: Mapping[str | Symbol, NumericishTypes]
    ) -> NumericishTypes:
        """Return a new expression with all the symbols substituted."""
        # Do a little data checky check
        if not all(symbol in self.symbols for symbol in substitutions):
            raise ValueError("Substitution contains symbols not in the expression")

        # Sort the substitutions into constants and expressions
        constants: Mapping[str, int | float | RangedValue] = {}
        constants_symbols = set()
        callables: Mapping[str, Expression | Symbol] = {}
        for symbol, value in substitutions.items():
            # FIXME: this is here because the context and expressions
            # hold reference to the values of attributes differently
            key = symbol.key if hasattr(symbol, "key") else symbol
            if callable(value):
                callables[key] = value
            else:
                constants[key] = value
                constants_symbols.add(symbol)

        # Create a new lambda function with the substitutions
        def _new_lambda(context):
            assert not (
                set(context) & set(constants)
            ), "Constants are being overwritten"
            new_context = {**context, **constants}
            for symbol, func in callables.items():
                new_context[symbol] = func(new_context)
            return self.lambda_(new_context)

        # Figure out what new symbols are required for this expression
        # Remove the constants we've substituted in, and add any new
        # symbols from the new expressions
        callables_symbols = set()
        for expr in callables.values():
            if isinstance(expr, Symbol):
                callables_symbols.add(expr)
                continue
            for symbol in expr.symbols:
                callables_symbols.add(symbol)
        new_symbols = self.symbols - set(substitutions.keys()) | callables_symbols

        # In the case we've completely substituted all the symbols
        # we can just return a new constant
        if not new_symbols:
            return _new_lambda({})

        return Expression(symbols=new_symbols, lambda_=_new_lambda)


def _get_symbols(thing: NumericishTypes) -> set[Symbol]:
    if isinstance(thing, Expression):
        return thing.symbols
    elif isinstance(thing, Symbol):
        return {thing}
    else:
        return set()


def defer_operation_factory(
    func: Callable,
    *args: NumericishTypes,
    src_ctx: ParserRuleContext | None = None,
) -> NumericishTypes:
    """Create a deferred operation, using deffering_type as the base for the callable."""
    if not any(map(callable, args)):
        # in this case we can just do the operation now, skip ahead and merry christmas
        try:
            return func(*args)
        except errors.AtoError as e:
            if src_ctx:
                e.set_src_from_ctx(src_ctx)
            raise

    # if we're here, we need to create an expression
    symbols = set.union(*map(_get_symbols, args))

    def _make_callable(not_callable):
        def _now_its_callable(_):
            return not_callable
        return _now_its_callable

    partials = [arg if callable(arg) else _make_callable(arg) for arg in args]

    def lambda_(context):
        try:
            return func(*(arg(context) for arg in partials))
        except errors.AtoError as e:
            if src_ctx:
                e.set_src_from_ctx(src_ctx)
            raise

    return Expression(symbols=symbols, lambda_=lambda_)


def simplify_expression_pool(
    pool: Mapping[str, NumericishTypes]
) -> Mapping[str, NumericishTypes]:
    """
    Simplify an expression, based on additional context we can give it.

    This is split from simplify_expression_pool, because we also need
    to simplify expressions that don't have symbols pointing to them.
    This is the case for the expressions in assertions, for example.

    # Usage

    The "pool" should contain a mapping of addresses to their assigned
    values. If something's declared, but not defined (eg. "A") in the
    below example, it should not be in the pool.

    # Problem Statement

    Take the example:
    A: int
    B = A + 1
    C = B + 2
    D = 3
    E = D + 4

    "A" remains a symbol, but we need to realise
    that it's a symbol, implying it's a strong, independent
    variable (who don't need no man)

    "B" is a simple expression, which we can't simplify,
    because it's already in terms of nothing other than constants
    and symbols.

    "C" is a more complex expression, so we can simplify it
    to "A + 3" or "(A + 1) + 2" for all I care.

    "D" is a constant, so it can't be simplified.

    "E" is an expression completely in terms of constants
    and defined symbols, so it can be simplified to "7".

    In summary:
    - Expressions or Symbols that are in terms of other
        expressions can be simplified
    - Expressions that are completely in terms of constants or
        defined symbols can be simplified
    - Nothing else can be simplified:
        - Constants and unassigned Symbols are independent, and
            can't be simplified
        - Expressions that are completely in independent terms
            can't be simplified further
    """

    # Expressions which we haven't got to yet
    touched = set()
    simplified = {}
    context = ChainMap(simplified, pool)

    def _visit(key: str, stack: list) -> NumericishTypes:
        if key in stack:
            loop = [key] + stack[:stack.index(key) + 1]
            loop = [address.get_instance_section(addr) for addr in loop]
            raise errors.AtoError(
                f"{' references '.join(loop)}",
                title="Circular dependency detected"
            )

        if key in touched:
            return context[key]  # no wakkas

        # Get the value from the pool
        value = context[key]

        # If this is something simple, just return it in the first place
        if not callable(value):
            touched.add(key)
            return value

        # If this thing points at something else, we can simplify it
        # For expressions, we find all the keys we have access to, and
        # go evaluate them first, before subbing them in
        if isinstance(value, Expression):
            simplified[key] = value.substitute(
                {
                    s: _visit(s.key, stack + [key])
                    for s in value.symbols
                    if s.key in pool
                }
            )
            touched.add(key)
            return simplified[key]

        # If it's a symbol, we simplify it by sticking the value in the
        # address there used to be a symbol instead
        elif isinstance(value, Symbol):
            if value.key in pool:
                simplified[key] = _visit(value.key, stack + [key])
                touched.add(key)
                return simplified[key]

            simplified[key] = value
            touched.add(key)
            return value

        raise ValueError("Unknown value type")

    # Iterate over the pool, simplifying as we go
    for key in pool:
        _visit(key, [])

    return simplified


def simplify_expression(
    expression: Expression,
    context: Mapping[str, NumericishTypes],
) -> Expression:
    """
    Simplify a single expression.
    This will only work if the context is already in its simplest forms.
    """
    substitutions = {symbol: context[symbol.key] for symbol in expression.symbols if symbol.key in context}
    expression = expression.substitute(substitutions)
    return expression



# %%
%load_ext autoreload
%autoreload 2
from atopile import api

# %%
model = api.build("/Users/mattwildoer/Projects/atopile-workspace/atopile/sandbox/test/elec/src/test.ato:Test")

# %%
print(model.a.current)

# %%



from atopile.address import AddrStr, get_parent_instance_addr, get_name, get_instance_section, get_relative_addr_str
from atopile.instance_methods import (
    get_children,
    get_links,
    get_supers_list,
    all_descendants,
    match_modules,
    match_components,
    match_interfaces,
    match_pins_and_signals,
)
import atopile.config

import networkx as nx

from collections import defaultdict
from typing import DefaultDict, Tuple


def get_parent(addr: AddrStr, build_ctx: atopile.config.BuildContext) -> AddrStr:
    """
    returns the parent of the given address or root if there is none
    """
    if addr == build_ctx.entry:
        return "null"
    elif get_parent_instance_addr(addr) == build_ctx.entry:
        return "root"

    return get_relative_addr_str(get_parent_instance_addr(addr), build_ctx.project_context.project_path)

def get_id(addr: AddrStr, build_ctx: atopile.config.BuildContext) -> AddrStr:
    """
    returns the parent of the given address or root if there is none
    """
    if addr == build_ctx.entry:
        return "root"

    return get_relative_addr_str(addr, build_ctx.project_context.project_path)

def get_blocks(addr, build_ctx: atopile.config.BuildContext) -> dict[str, dict[str, str]]:
    """
    returns a dictionary of blocks:
    {
        "block_name": {
            "instance_of": "instance_name",
            "type": "module/component/interface/signal/builtin",
            "address": "a.b.c",
            "value": "value"
        }, ...
    }
    """
    block_dict = {}
    for child in get_children(addr):
        if match_modules(child) or match_components(child) or match_interfaces(child) or match_pins_and_signals(child):
            type = "module"
            if match_components(child):
                type = "component"
            elif match_interfaces(child):
                type = "interface"
            elif match_pins_and_signals(child):
                type = "signal"
            else:
                type = "module"

            block_dict[get_name(child)] = {
                "instance_of": get_name(get_supers_list(child)[0].obj_def.address),
                "type": type,
                "address": get_relative_addr_str(child, build_ctx.project_context.project_path)}

    return block_dict

def process_links(addr: AddrStr) -> list[dict]:
    """
    returns a list of links:
    [
        {
            "source": {
                "block": "block_name",
                "port": "port_name"
            },
            "target": {
                "block": "block_name",
                "port": "port_name"
            },
            "type": "interface/signal"
        }, ...
    ]
    """
    link_list = []
    links = get_links(addr)
    for link in links:
        # Type is either interface or signal
        if match_pins_and_signals(link.source.addr):
            type = "signal"
            instance_of = "signal"
        else:
            type = "interface"
            instance_of = get_name(get_supers_list(link.source.addr)[0].obj_def.address)
        source_block, source_port = split_list_at_n(get_current_depth(addr), split_addr(link.source.addr))
        target_block, target_port = split_list_at_n(get_current_depth(addr), split_addr(link.target.addr))

        _source = {"block": get_name(combine_addr(source_block)), "port": combine_addr(source_port)}
        _target = {"block": get_name(combine_addr(target_block)), "port": combine_addr(target_port)}
        link_list.append({"source": _source, "target": _target, "type": type, "instance_of": instance_of})

    return link_list


def is_path_without_end_nodes(G, blocks, source, target):
    """"
    Is there a direct path between two nodes? (allowed to go through chained signals and interfaces)
    """
    #TODO: Got tired of finding nets myself so started using networkx.
    # Probably going to have to cluster all those utils together somewhere
    if source == target or source not in G or target not in G:
        return False
    # Remove other end nodes from the graph temporarily
    G_temp = G.copy()
    for block in blocks:
        if block not in (source, target) and (blocks[block]['type'] == "module" or blocks[block]['type'] == "component") and block in G_temp.nodes:
            G_temp.remove_node(block)
    # Check if there's a path in the modified graph
    has_path = nx.has_path(G_temp, source, target)

    return has_path

# Deprecated but keeping here for reference
def get_block_to_block_links(addr: AddrStr) -> list[tuple[str, str]]:
    """
    returns a list of block to block links:
    [
        {'source': 'block_name', 'target': 'block_name'}, ...
    ]
    """
    blocks = get_blocks(addr)
    links = process_links(addr)

    # Create a graph with the blocks
    G = nx.Graph()
    #G.add_nodes_from(blocks)

    # Add the links to the graph
    for link in links:
        if link['type'] == "interface" and link['instance_of'] != "Power":
            G.add_edge(link['source']['block'], link['target']['block'])
        # G.add_edge(link['source']['block'], link['target']['block'])

    # Identify the block-to-block connections
    block_connections = set()
    for source in blocks:
        if blocks[source]['type'] == "module" or blocks[source]['type'] == "component":
            for target in blocks:
                if blocks[target]['type'] == "module" or blocks[target]['type'] == "component":
                    if source != target and is_path_without_end_nodes(G, blocks, source, target):
                        block_connections.add((source, target))

    unique_tuples_list = {tuple(sorted(t)) for t in block_connections}

    block_to_block_list = []
    for connection in unique_tuples_list:
        block_to_block_list.append({'source': connection[0], 'target': connection[1]})

    return block_to_block_list

# Bundled connections will be called harnesses
def get_harnesses(addr: AddrStr) -> list[dict]:
    """
    returns a list of bundled connections (harnesses):
    [
        {
            "source": "block_name",
            "target": "block_name",
            "name": "name_1/name_2/...",
            "preview_names": ["name_1", "name_2",...],
            "links": [
                {
                    "source": "port_name",
                    "target": "port_name",
                    "type": "interface/signal"
                    "instance_of": "instance_name"
                }
        }, ...
    ]
    """
    harness_list: DefaultDict[Tuple[str, str], list] = defaultdict(list)
    links = get_links(addr)
    for link in links:
        # Type is either interface or signal
        if match_pins_and_signals(link.source.addr):
            type = "signal"
            instance_of = "signal"
        else:
            type = "interface"
            instance_of = get_name(get_supers_list(link.source.addr)[0].obj_def.address)
        source_block, source_port = split_list_at_n(get_current_depth(addr), split_addr(link.source.addr))
        target_block, target_port = split_list_at_n(get_current_depth(addr), split_addr(link.target.addr))

        source_block = get_name(combine_addr(source_block))
        target_block = get_name(combine_addr(target_block))
        source_port = combine_addr(source_port)
        target_port = combine_addr(target_port)

        harness_list[(source_block, target_block)].append({
            "source": source_port,
            "target": target_port,
            "type": type,
            "instance_of": instance_of
        })

    harness_return_dict: DefaultDict[str, list] = defaultdict(list)

    for source_block, target_block in harness_list:
        preview_name_set = set()
        for link in harness_list[(source_block, target_block)]:
            if link['type'] == "signal":
                preview_name_set.add(f"{link['source']} ~ {link['target']}")
            # special case for pairs as they could be anything
            elif link['type'] == "interface" and link['instance_of'] == "Pair":
                preview_name_set.add(f"{link['source']} ~ {link['target']}")
            # for interfaces
            else:
                preview_name_set.add(link['instance_of'])

        preview_names = sorted(preview_name_set)
        name = "/".join(preview_names)
        key = f"{source_block}_{target_block}"
        harness_return_dict[key] = {
            "source": source_block,
            "target": target_block,
            "name": name,
            "preview_names": preview_names,
            "links": harness_list[(source_block, target_block)]
        }

    return harness_return_dict


def get_vis_dict(build_ctx: atopile.config.BuildContext) -> dict:
    return_json = {}
    # for addr in chain(root, all_descendants(root)):
    for addr in all_descendants(build_ctx.entry):
        block_dict = {}
        link_list = []
        # we only create an entry for modules, not for components
        if match_modules(addr) and not match_components(addr):
            instance = get_id(addr, build_ctx)
            parent = get_parent(addr, build_ctx)
            block_dict = get_blocks(addr, build_ctx)
            link_list = process_links(addr)
            harness_dict = get_harnesses(addr)

            return_json[instance] = {
                "parent": parent,
                "blocks": block_dict,
                "links": link_list,
                "harnesses": harness_dict,
            }

    return return_json

def get_current_depth(addr: AddrStr) -> int:
    instance_section = get_instance_section(addr)
    if instance_section is None or instance_section == "":
        return 0
    else:
        return len(instance_section.split("."))

def split_addr(addr: AddrStr) -> list[str]:
    return get_instance_section(addr).split(".")

def combine_addr(list: list[str]) -> str:
    return ".".join(list)

def split_list_at_n(n, list_of_strings):
    # Split the list
    first_part = list_of_strings[:n+1]
    second_part = list_of_strings[n+1:]

    return first_part, second_part



import textwrap
from pathlib import Path

from atopile import front_end, parse
from atopile.front_end import parser

PRJ = Path(__file__).parent / "prj"
FILE = PRJ / "test.ato"
MODULE = str(FILE) + ":Test"


def test_caching():
    parser.cache[str(FILE)] = parse.parse_text_as_file(
        textwrap.dedent(
            """
            module Test:
                a = 1
                b = 2
            """
        ),
        MODULE,
    )
    root_1 = front_end.lofty.get_instance(MODULE)

    # Make sure it's parsed properly
    assert root_1.assignments["a"][0].value == 1

    # Make sure it's consistent and cached
    assert root_1 is front_end.lofty.get_instance(MODULE)

    # Clear it out, to make sure it's re-parsed
    front_end.reset_caches(MODULE)

    # Parse something else to make sure it's re-parsed
    root_2 = front_end.lofty.get_instance(MODULE)
    assert root_1 is not root_2



# pylint: disable=logging-fstring-interpolation

"""
`ato view`
"""
import logging
import textwrap
from enum import Enum
from pathlib import Path

import click
import yaml
from quart import Quart, jsonify, send_from_directory
from quart_cors import cors
from quart_schema import QuartSchema, validate_request, validate_response
from watchfiles import awatch

import atopile.address
import atopile.config
import atopile.front_end
import atopile.instance_methods
import atopile.schematic_utils
import atopile.viewer_utils
from atopile import errors
from atopile.cli.common import project_options
from atopile.config import BuildContext, set_project_context
from atopile.viewer_core import Pose

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


app = Quart(__name__, static_folder="../viewer/dist", static_url_path="")
app = cors(app, allow_origin="*")
QuartSchema(app)


@app.route("/block-diagram-data")
async def send_viewer_data():
    build_ctx: BuildContext = app.config["build_ctx"]
    return jsonify(atopile.viewer_utils.get_vis_dict(build_ctx))


class DiagramType(str, Enum):
    """The type of diagram."""
    block = "block"
    schematic = "schematic"


@app.route("/<diagram_type>/<path:addr>/pose", methods=["POST"])
@validate_request(Pose)
@validate_response(Pose, 201)
async def save_pose(
    diagram_type: str | DiagramType,
    addr: str,
    data: Pose
) -> tuple[Pose, int]:
    """Save the pose of an element."""
    diagram_type = DiagramType(diagram_type)
    build_ctx: BuildContext = app.config["build_ctx"]

    # FIXME: rip this logic outta here
    # We save the pose information to one file per-project
    # FIXME: figure out how we should actually
    # interact with these config files
    lock_path = build_ctx.project_context.lock_file_path
    if lock_path.exists():
        with lock_path.open("r") as lock_file:
            lock_data = yaml.safe_load(lock_file) or {}
    else:
        lock_data = {}

    lock_data.setdefault("poses", {}).setdefault(diagram_type.name, {})[addr] = data.model_dump()

    with lock_path.open("w") as lock_file:
        yaml.safe_dump(lock_data, lock_file)

    return data, 200


@app.route("/schematic-data")
async def send_schematic_data():
    build_ctx: BuildContext = app.config["build_ctx"]
    return jsonify(atopile.schematic_utils.get_schematic_dict(build_ctx))


@app.route("/")
async def home():
    """Serve the home page."""
    return await send_from_directory(app.static_folder, "index.html")


def _ato_file_filter(_, path: str):
    """Filter for files that are not ato files."""
    return path.endswith(".ato")


async def monitor_changes(src_dir: Path):
    """Background task to monitor the project for changes."""
    log.info(f"Monitoring {src_dir} for changes")

    async for changes in awatch(src_dir, watch_filter=_ato_file_filter, recursive=True):
        for change, file in changes:
            log.log(logging.NOTSET, "Change detected in %s: %s", file, change.name)
            atopile.front_end.reset_caches(file)


@app.before_serving
async def startup():
    """Set up the viewer."""
    log.info("Setting up the viewer")
    build_ctx: BuildContext = app.config["build_ctx"]

    # Monitor the project for changes to reset caches
    app.add_background_task(
        monitor_changes,
        build_ctx.project_context.project_path
    )

    # Pre-build the entry point
    atopile.instance_methods.get_instance(build_ctx.entry)


@click.command()
@project_options
def view(build_ctxs: list[BuildContext]):
    """
    View a block diagram or schematic of your project.
    """
    log.info("Spinning up the viewer")

    if len(build_ctxs) != 1:
        log.info(textwrap.dedent("""
            You need to select what you want to view.
            - If you use the `--build` option, you will view the entry point of the build.
            - If you add an argument for the address, you'll view that.
        """))
        raise errors.AtoNotImplementedError("Multiple build configs not yet supported.")

    build_ctx = build_ctxs[0]
    app.config["build_ctx"] = build_ctx
    set_project_context(build_ctx.project_context)

    app.run(host="127.0.0.1", port=8080)



import re
import sys
from typing import TextIO, Any

from antlr4 import InputStream, Lexer
from antlr4.Token import CommonToken

from .AtopileParser import AtopileParser

import atopile.errors


class AtopileLexerBase(Lexer):
    NEW_LINE_PATTERN = re.compile('[^\r\n\f]+')
    SPACES_PATTERN = re.compile('[\r\n\f]+')

    def __init__(self, input: InputStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.tokens = []
        self.indents = []
        self.opened = 0  # "opened" indicates braces/brackets etc...
        self.comments: dict[tuple[Any, int], str] = {}

    def reset(self):
        self.tokens = []
        self.indents = []
        self.opened = 0
        self.comments = {}
        super().reset()

    def emitToken(self, token):
        self._token = token
        self.tokens.append(token)

    def nextToken(self):
        # Check if the end-of-file is ahead and there are still some DEDENTS expected.
        if self._input.LA(1) == AtopileParser.EOF and len(self.indents) != 0:
            # Remove any trailing EOF tokens from our buffer.
            self.tokens = [token for token in self.tokens if token.type != AtopileParser.EOF]

            # First emit an extra line break that serves as the end of the statement.
            self.emitToken(self.commonToken(AtopileParser.NEWLINE, '\n'))

            # Now emit as much DEDENT tokens as needed.
            while len(self.indents) != 0:
                self.emitToken(self.createDedent())
                self.indents.pop()

            # Put the EOF back on the token stream.
            self.emitToken(self.commonToken(AtopileParser.EOF, '<EOF>'))

        next_ = super().nextToken()
        return next_ if len(self.tokens) == 0 else self.tokens.pop(0)

    def createDedent(self):
        return self.commonToken(AtopileParser.DEDENT, '')

    def commonToken(self, type_: int, text: str):
        stop = self.getCharIndex() - 1
        start = stop if text == '' else stop - len(text) + 1
        return CommonToken(self._tokenFactorySourcePair, type_, Lexer.DEFAULT_TOKEN_CHANNEL, start, stop)

    def getIndentationCount(self, whitespace: str):
        count = 0
        for c in whitespace:
            if c == '\t':
                raise atopile.errors.AtoNotImplementedError(
                    "Tabs aren't supported. Please use spaces instead",
                    src_path=self.inputStream.name,
                    src_line=self._tokenStartLine,
                    src_col=self._tokenStartColumn
                )
                # FIXME: this should be based on the last indent or something else instead of 8
                count += 8 - count % 8
            else:
                count += 1
        return count

    def atStartOfInput(self):
        return self.getCharIndex() == 0

    def openBrace(self):
        self.opened += 1

    def closeBrace(self):
        self.opened -= 1

    def onNewLine(self):
        new_line = self.NEW_LINE_PATTERN.sub('', self.text)
        spaces = self.SPACES_PATTERN.sub('', self.text)

        # Strip newlines inside open clauses except if we are near EOF. We keep NEWLINEs near EOF to
        # satisfy the final newline needed by the single_put rule used by the REPL.
        next_ = self._input.LA(1)
        next_next = self._input.LA(2)

        # unicode 10 = line-feed
        # unicode 13 = carriage-return
        # unicode 35 = hash (comments)
        if self.opened > 0 or (next_next != -1 and next_ in (10, 13, 35)):
            self.skip()

        else:
            self.emitToken(self.commonToken(AtopileParser.NEWLINE, new_line))
            indent = self.getIndentationCount(spaces)
            previous = 0 if len(self.indents) == 0 else self.indents[-1]

            if indent == previous:
                self.skip()
            elif indent > previous:
                self.indents.append(indent)
                self.emitToken(self.commonToken(AtopileParser.INDENT, spaces))
            else:
                while len(self.indents) > 0 and self.indents[-1] > indent:
                    self.emitToken(self.createDedent())
                    self.indents.pop()

    def skip(self):
        """
        Skip the token, but in case it's a comment,
        first store it in the comments dictionary.
        """
        # FIXME: there's surely a better way to distinguish comments
        if self.text[0] == "#":
            comment = self.text[1:].strip()
            self.comments[(self.inputStream.name, self._tokenStartLine)] = comment
        super().skip()



import pint

from atopile.expressions import RangedValue


def test_comparitors():
    a = RangedValue(1, 2)  # a: |---|
    b = RangedValue(2, 3)  # b:     |---|
    c = RangedValue(1, 3)  # c: |-------|
    d = RangedValue(3, 4)  # d:         |---|

    assert a <= b
    assert not a < b
    assert a < d

    assert b >= a
    assert not b > a

    assert not a.within(b)
    assert a.within(c)
    assert a == RangedValue(1, 2)


def test_units():
    a = RangedValue(1000, 2000, "mV")
    b = RangedValue(1, 2, "kV")
    assert a < b

    c = a * b
    assert c.unit == pint.Unit("V") ** 2


def test_arithmetic():
    assert RangedValue(1, 2) + RangedValue(2, 3) == RangedValue(3, 5)
    assert RangedValue(1, 2) - RangedValue(2, 3) == RangedValue(-2, 0)
    assert RangedValue(1, 2) * RangedValue(2, 3) == RangedValue(2, 6)
    assert RangedValue(1, 2) / RangedValue(2, 3) == RangedValue(1 / 3, 1)
    assert RangedValue(1, 2) ** 2 == RangedValue(1, 4)

    assert RangedValue(1, 2) + 2 == RangedValue(3, 4)
    assert RangedValue(1, 2) - 3 == RangedValue(-2, -1)
    assert RangedValue(1, 2) * 4 == RangedValue(4, 8)
    assert RangedValue(1, 2) / 5 == RangedValue(1 / 5, 2 / 5)

    assert 2 + RangedValue(1, 2) == RangedValue(3, 4)
    assert 3 - RangedValue(1, 2) == RangedValue(-2, -1)
    assert 4 * RangedValue(1, 2) == RangedValue(4, 8)
    assert 5 / RangedValue(1, 2) == RangedValue(5 / 2, 5)


def test_pretty_str():
    str_rep = "mhmm, this isn't a ranged value"
    assert RangedValue(3, 3, pint.Unit("V"), str_rep=str_rep).pretty_str() == str_rep
    assert RangedValue(3, 3, pint.Unit("V")).pretty_str() == "3V"
    assert RangedValue(3, 5, pint.Unit("V")).pretty_str() == "3 to 5 V"
    assert RangedValue(3, 5, pint.Unit("mV")).pretty_str() == "3 to 5 mV"
    assert RangedValue(3, 3.1, pint.Unit("V")).pretty_str() == "3.05V ± 50mV"

    # # Test Zeros
    assert RangedValue(0, 0, pint.Unit("V")).pretty_str() == "0V"
    assert RangedValue(0, 50, pint.Unit("V")).pretty_str() == "0 to 50 V"
    assert RangedValue(-50, 0, pint.Unit("V")).pretty_str() == "-50 to 0 V"

    # Test Negatives
    assert RangedValue(-50, -48, pint.Unit("V")).pretty_str() == "-49 ± 1 V"
    assert RangedValue(-50, -30, pint.Unit("V")).pretty_str() == "-50 to -30 V"
    assert RangedValue(-50, 50, pint.Unit("V")).pretty_str() == "± 50V"
    assert RangedValue(-50, 50, pint.Unit("mV")).pretty_str() == "± 50mV"

    # Make sure combined units compact properly
    v = RangedValue(3, 3, pint.Unit("V"))
    r = RangedValue(1000, 1000, pint.Unit("Ω"))
    i = v / r
    assert i.pretty_str() == "3mA"



import logging

from atopile.address import AddrStr, get_name, add_instance, get_entry_section, get_relative_addr_str
from atopile.instance_methods import (
    get_children,
    get_links,
    get_supers_list,
    all_descendants,
    get_parent,
    match_modules,
    match_components,
    match_interfaces,
    match_signals,
    match_pins_and_signals
)
from atopile.front_end import Link
from atopile import errors
import atopile.config
from atopile.viewer_core import Pose, Position

from atopile.viewer_utils import get_id

import json
import yaml

from typing import Optional

import hashlib

from atopile.components import get_specd_value, MissingData

log = logging.getLogger(__name__)


_get_specd_value = errors.downgrade(get_specd_value, MissingData)

#FIXME: this function is a reimplementation of the one in instance methods, since I don't have access to the std lib
# Diff is the additon of get_name(...)
def find_matching_super(
    addr: AddrStr, candidate_supers: list[AddrStr]
) -> Optional[AddrStr]:
    """
    Return the first super of addr, is in the list of
    candidate_supers, or None if none are.
    """
    supers = get_supers_list(addr)
    for duper in supers:
        if any(get_name(duper.address) == pt for pt in candidate_supers):
            return duper.address
    return None

def get_std_lib(addr: AddrStr) -> str:
    #TODO: The time has come to bake the standard lib as a compiler dependency...
    std_lib_supers = [
        "Resistor",
        "Inductor",
        "Capacitor",
        "LED",
        "Power",
        "NPN",
        "PNP",
        "Diode",
        "SchottkyDiode",
        "ZenerDiode",
        "NFET",
        "PFET",
        "Opamp"]

    # Handle signals
    if match_signals(addr):
        signal_parent = get_parent(addr)
        if match_interfaces(signal_parent):
            matching_super = find_matching_super(signal_parent, std_lib_supers)
            if matching_super is not None:
                if get_entry_section(matching_super) == "Power":
                    return "Power." + get_name(addr)
            else:
                return ""

    # handle components
    matching_super = find_matching_super(addr, std_lib_supers)
    if matching_super is None:
        return ""
    return get_name(matching_super)

def get_schematic_dict(build_ctx: atopile.config.BuildContext) -> dict:
    return_json: dict = {}

    ato_lock_contents = get_ato_lock_file(build_ctx)

    for addr in all_descendants(build_ctx.entry):
        if match_modules(addr) and not match_components(addr):
            # Start by creating a list of all the components in the view and their ports.
            # Ports consist of cluster of pins, signals and signals within interfaces

            # Component dict that we will return
            components_dict: dict[AddrStr, dict] = {}

            # Those are all the modules at or below the module currently being inspected
            blocks_at_and_below_view: list[AddrStr] = list(filter(match_modules, all_descendants(addr)))
            blocks_at_and_below_view.extend(list(filter(match_interfaces, all_descendants(addr))))

            links_at_and_below_view: list[Link] = []
            for module in blocks_at_and_below_view:
                links_at_and_below_view.extend(list(get_links(module)))

            pins_and_signals_at_and_below_view = list(filter(match_pins_and_signals, all_descendants(addr)))

            # This is a map of connectables beneath components to their net cluster id
            connectable_to_nets_map: dict[AddrStr, str] = {}

            signals_dict: dict[AddrStr, dict] = {}

            # We start exploring the modules
            for block in blocks_at_and_below_view:
                #TODO: provide error message if we can't handle the component
                if match_components(block):
                    component = block
                    # There might be nested interfaces that we need to extract
                    blocks_at_or_below_component = list(filter(match_modules, all_descendants(component)))
                    # Extract all links at or below the current component and form nets
                    links_at_and_below_component = []
                    for block in blocks_at_or_below_component:
                        links_at_and_below_component.extend(list(get_links(block)))

                    pins_and_signals_at_and_below_component = list(filter(match_pins_and_signals, all_descendants(component)))
                    component_nets = find_nets(pins_and_signals_at_and_below_component, links_at_and_below_component)

                    pose = Pose()

                    # Component ports
                    component_ports_dict: dict[int, dict[str, str]] = {}
                    for component_net_index, component_net in enumerate(component_nets):
                        # create a hash of the net
                        hash_object = hashlib.sha256()
                        json_string = json.dumps(component_net)
                        hash_object.update(json_string.encode())
                        net_hash = hash_object.hexdigest()[:8]

                        pose = get_pose(ato_lock_contents, net_hash)

                        component_ports_dict[component_net_index] = {
                            "net_id": net_hash,
                            "name": '/'.join(map(get_name, component_net)),
                            "position": pose.position,
                            "rotation": pose.rotation,
                            "mirror_x": pose.mirror_x,
                            "mirror_y": pose.mirror_y
                        }

                        for connectable in component_net:
                            connectable_to_nets_map[connectable] = net_hash

                    comp_addr = get_relative_addr_str(component, build_ctx.project_context.project_path)
                    pose = get_pose(ato_lock_contents, comp_addr)

                    components_dict[comp_addr] = {
                        "instance_of": get_name(get_supers_list(component)[0].obj_def.address),
                        "std_lib_id": get_std_lib(component),
                        "value": _get_specd_value(component),
                        "address": get_relative_addr_str(component, build_ctx.project_context.project_path),
                        "name": get_name(component),
                        "ports": component_ports_dict,
                        "position": pose.position,
                        "rotation": pose.rotation,
                        "mirror_x": pose.mirror_x,
                        "mirror_y": pose.mirror_y
                    }

                elif match_interfaces(block):
                    pass

                else:
                    #TODO: this only handles interfaces in the highest module, not in nested modules
                    interfaces_at_module = list(filter(match_interfaces, get_children(block)))
                    for interface in interfaces_at_module:
                        #TODO: handle signals or interfaces that are not power
                        signals_in_interface = list(filter(match_signals, get_children(interface)))
                        for signal in signals_in_interface:
                            signals_dict[signal] = {
                                "std_lib_id": get_std_lib(signal),
                                "instance_of": get_name(get_supers_list(interface)[0].obj_def.address),
                                "address": get_relative_addr_str(signal, build_ctx.project_context.project_path),
                                "name": get_name(get_parent(signal)) + "." + get_name(signal)}

                            if get_std_lib(signal) != "none":
                                pass
                                #connectable_to_nets_map[signal] = signal

                    signals_at_view = list(filter(match_signals, get_children(block)))
                    for signal in signals_at_view:
                        signals_dict[signal] = {
                            "std_lib_id": get_std_lib(signal),
                            "instance_of": "signal",
                            "address": get_relative_addr_str(signal, build_ctx.project_context.project_path),
                            "name": get_name(signal)}



            # This step is meant to remove the irrelevant signals and interfaces so that we
            # don't show them in the viewer
            nets_above_components = find_nets(pins_and_signals_at_and_below_view, links_at_and_below_view)
            converted_nets_above_components = []
            for net in nets_above_components:
                # Make it a set so we don't add multiple times the same hash
                converted_net = set()
                for connectable in net:
                    if connectable in connectable_to_nets_map:
                        converted_net.add(connectable_to_nets_map[connectable])
                converted_nets_above_components.append(list(converted_net))

            # net_links = []
            # # for each net in the component_net_above_components, create a link between each of the nodes in the net
            # for net in converted_nets_above_components:
            #     output_net = []
            #     for conn in net:
            #         output_net.append(conn)
            #     net_links.append(output_net)

            instance = get_id(addr, build_ctx)
            return_json[instance] = {
                "components": components_dict,
                "signals": signals_dict,
                "nets": converted_nets_above_components
            }

    return return_json

def get_ato_lock_file(build_ctx: atopile.config.BuildContext) -> dict:
    ato_lock_contents = {}

    if build_ctx.project_context.lock_file_path.exists():
        with build_ctx.project_context.lock_file_path.open("r") as lock_file:
            ato_lock_contents = yaml.safe_load(lock_file)

    return ato_lock_contents

def get_pose(ato_lock_contents: dict, id: str) -> Pose:
    position = ato_lock_contents.get("poses", {}).get("schematic", {}).get(id, {}).get("position", {'x': 0, 'y': 0})
    rotation = ato_lock_contents.get("poses", {}).get("schematic", {}).get(id, {}).get("rotation", 0)
    mirror_x = ato_lock_contents.get("poses", {}).get("schematic", {}).get(id, {}).get("mirror_x", False)
    mirror_y = ato_lock_contents.get("poses", {}).get("schematic", {}).get(id, {}).get("mirror_y", False)

    return Pose(
        position=Position(x=position['x'], y=position['y']),
        rotation=rotation,
        mirror_x=mirror_x,
        mirror_y=mirror_y
    )

#TODO: copied over from `ato inspect`. We probably need to deprecate `ato inspect` anyways and move this function
# to a common location
def find_nets(pins_and_signals: list[AddrStr], links: list[Link]) -> list[list[AddrStr]]:
    """
    pins_and_signals: list of all the pins and signals that are expected to end up in the net
    links: links that connect the pins_and_signals_together
    """
    # Convert links to an adjacency list
    graph = {}
    for pin_and_signal in pins_and_signals:
        graph[pin_and_signal] = []

    # Short the pins and signals to each other on the first run to make sure they are recorded as nets
    for link in links:
        source = []
        target = []
        if match_interfaces(link.source.addr) and match_interfaces(link.target.addr):
            for int_pin in get_children(link.source.addr):
                if match_pins_and_signals(int_pin):
                    source.append(int_pin)
                    target.append(add_instance(link.target.addr, get_name(int_pin)))
                else:
                    raise errors.AtoNotImplementedError("Cannot nest interfaces yet.")
        elif match_interfaces(link.source.addr) or match_interfaces(link.target.addr):
            # If only one of the nodes is an interface, then we need to throw an error
            raise errors.AtoTypeError.from_ctx(
                link.src_ctx,
                f"Cannot connect an interface to a non-interface: {link.source.addr} ~ {link.target.addr}"
            )
        # just a single link
        else:
            source.append(link.source.addr)
            target.append(link.target.addr)

        for source, target in zip(source, target):
            if source not in graph:
                graph[source] = []
            if target not in graph:
                graph[target] = []
            graph[source].append(target)
            graph[target].append(source)


    def dfs(node, component):
        visited.add(node)
        component.append(node)
        for neighbor in graph.get(node, []):
            if neighbor not in visited:
                dfs(neighbor, component)

    connected_components = []
    visited = set()
    for node in graph:
        if node not in visited:
            component = []
            dfs(node, component)
            connected_components.append(component)

    return connected_components